/**************************************************************************/
/*									  */
/*	(c) 1994 BinTec Computersysteme GmbH				  */
/*	All Rights Reserved						  */
/*									  */
/*									  */
/*									  */
/* Permission to use, copy, modify, and distribute this software and its  */
/* documentation for any purpose and without fee is hereby granted.	  */
/* This software is provided "as is" without express or implied warranty. */
/*									  */
/**************************************************************************/

/*
	SFF fax file format is used by the BinTec BIANCA/ISDN
	FAX-G3 driver software.
*/


/**************************************************************************

	changes to the Ghostscript 2.6.1 Makefile
        =========================================

DEVICE_DEVS=sfflow.dev sffhigh.dev



### ------------------------- SFF file formats ------------------------- ###
 
sff_=gdevsff.$(OBJ) gdevpccm.$(OBJ) gdevprn.$(OBJ)
 
gdevsff.$(OBJ): gdevsff.c $(PDEVH) $(gdevpccm_h)
 
sfflow.dev: $(sff_)
        $(SHP)gssetdev sfflow $(sff_)
 
sffhigh.dev: $(sff_)
        $(SHP)gssetdev sffhigh $(sff_)

**************************************************************************/




/* gdevsff.c */
/* SFF file format devices for Ghostscript */
/* JIM */

#include "gdevprn.h"
#include "faxmh.h"	


struct gx_device_sff_s {
	gx_device_common;
	gx_prn_device_common;
};
typedef struct gx_device_sff_s gx_device_sff;


private dev_proc_open_device(sff_open);
private dev_proc_print_page(sff_printhigh_page);
private dev_proc_print_page(sff_printlow_page);
private dev_proc_close_device(sff_close);

gx_device_procs sff_std_procs =
    prn_procs( sff_open, gdev_prn_output_page, sff_close);

gx_device_sff far_data gs_sffhigh_device =
{   prn_device_std_body(
        gx_device_sff,
        sff_std_procs,
        "sffhigh",
        90,
        115,
        204, 196,
        0,0,0,0,
        1,
        sff_printhigh_page
    )
};

gx_device_sff far_data gs_sfflow_device =
{   prn_device_std_body(
        gx_device_sff,
        sff_std_procs,
        "sfflow",
        90,
        115,
        204, 98,
        0,0,0,0,
        1,
        sff_printlow_page
    )
};




/* ------ Private definitions ------ */

static long fax_pixels;
static long fax_rempix;
static unsigned char fax_buffer[1500];
static unsigned int  fax_index;
static unsigned fax_pages;
static FILE *fax_fp;

static fax_char( c, file )
unsigned char c;
FILE *file;
{
    if (fputc( c & 0xff, file) == EOF) {
	fprintf( stderr, "fax_char() fputc failed\n");
    }
}

static fax_short( c, file )
unsigned short c;
FILE *file;
{
    fax_char( c & 0xff, file);
    fax_char( c >> 8, file);
}

static fax_long( c, file )
unsigned long c;
FILE *file;
{
    fax_char( c & 0xff, file);
    fax_char( c >> 8, file);
    fax_char( c >> 16, file);
    fax_char( c >> 24, file);
}

static void fax_sffdoc(file)
FILE *file;
{
    fax_char( 'S', file); 
    fax_char( 'f', file); 
    fax_char( 'f', file); 
    fax_char( 'f', file);
    fax_char(  1,  file);
    fax_char(  0,  file);
    fax_short( 0,  file);
    fax_short( 0,  file);
    fax_short(20,  file);
    fax_long(  0,  file);
    fax_long(  0,  file);
}
 
static void fax_sffpage(pdev, res, file)
gx_device_printer *pdev; 
int res;
FILE *file;
{
    fax_char(   254, file);
    fax_char(    16, file);
    fax_char(   res, file);
    fax_char(     0, file);
    fax_char(     0, file);
    fax_char(     0, file);
    fax_short( 1728, file);
    fax_short( pdev->height, file);
    fax_long(     0, file);
    fax_long(     0, file);
}

static void fax_sffend( file )
FILE *file;
{
    fax_char( 254, file);
    fax_char(   0, file);
}


static void sff_write_line( from, end, file)
unsigned char *from; 
unsigned char *end; 
FILE *file;
{	
    short i;

    short color  = 0;
    short cnt    = 0;

    fax_rempix = 0;
    fax_pixels = 0;
    fax_index  = 0;

    for (i = 0; i < (216*8); ++i) {
	if (color) {
	    if ( from[i/8] & (0x80 >> (i%8))) {
		++cnt;
	    } else {
		fax_t4emit( 1, cnt, 0);
		color = 0;
		cnt   = 1;
	    }
	} else {
	    if ( from[i/8] & (0x80 >> (i%8))) {
		fax_t4emit( 0, cnt, 0);
		color = 1;
		cnt   = 1;
	    } else {
		++cnt;
	    }
	}
    }
    fax_t4emit( color, cnt, 1);

    if (fax_index <= 216) {
	fax_char( fax_index, file );
    } else {
	fax_char (0, file);
	fax_short(fax_index, file);
    }
    for (i = 0; i < fax_index; ++i) fax_char(fax_buffer[i], file);
}

static int fax_t4sendbits( p, len, f)
int p, len, f;
{
    fax_pixels |= p << fax_rempix;
    fax_rempix += len;
    while (fax_rempix >= 8) {
	if (fax_index < sizeof(fax_buffer))
	    fax_buffer[fax_index++] = fax_pixels;
        fax_rempix -= 8;
        fax_pixels >>= 8;
    }
    if (f && fax_rempix) {
        fax_rempix = 0;
	if (fax_index < sizeof(fax_buffer))
	    fax_buffer[fax_index++] = fax_pixels;
    }
}

static int fax_t4emit(black, cnt, f)
int black, cnt, f;
{
    int i;

    i = cnt >> 6;
    cnt &= 0x3f;
    if (black) {
        if (i) {
            fax_t4sendbits( __mc_black[i-1].pixels, __mc_black[i-1].len, 0);
        }
        fax_t4sendbits(__tc_black[cnt].pixels, __tc_black[cnt].len, f);
    } else {
        if (i) {
            fax_t4sendbits( __mc_white[i-1].pixels, __mc_white[i-1].len, 0);
        }
        fax_t4sendbits( __tc_white[cnt].pixels, __tc_white[cnt].len, f);
    }
}

private int sff_write_page( pdev, res, file)
gx_device_printer *pdev; 
int res;
FILE *file; 
{	
    int raster  = gdev_prn_raster(pdev);
    int height  = pdev->height;
    uint rsize  = (pdev->width + 7) >> 3;
    byte *row   = (byte *)gs_malloc(raster + rsize, 1, "sff file buffer");
    byte *end   = row + raster;
    int y;

    if (!row)
	return_error(gs_error_VMerror);

    fax_pages++;
    fax_sffpage(pdev, res, file);

    for ( y = 0; y < height; y++ ) {	
	gdev_prn_copy_scan_lines(pdev, y, row, raster);
	sff_write_line(row, end, file);
    }

pcx_done:
    gs_free((char *)row, raster + rsize, 1, "sff file buffer");

    return 0;
}

private int sff_printhigh_page( pdev, file)
gx_device_printer *pdev; 
FILE *file;
{
    if (!fax_fp) fax_fp = file;
    if (!fax_pages) fax_sffdoc(file);
    return sff_write_page(pdev, 1, file);
}

private int sff_printlow_page( pdev, file)
gx_device_printer *pdev; 
FILE *file;
{
    if (!fax_fp) fax_fp = file;
    if (!fax_pages) fax_sffdoc(file);
    return sff_write_page(pdev, 0, file);
}

private int sff_open( pdev )
gx_device *pdev;
{
    int rc;
    fax_pages = 0;
    fax_fp = 0;
    rc = gdev_prn_open(pdev);
    return (rc);
}

private int sff_close( pdev )
gx_device *pdev;
{
    if (fax_fp) {
	fax_sffend( fax_fp);
	fflush(fax_fp);
	if (fseek(fax_fp, 8, 0) != -1) {
	    fax_short( fax_pages, fax_fp);
	}

    }
    return (gdev_prn_close(pdev));
}



